import traceback

from datetime import datetime
from logging import getLogger
from typing import Final


__all__ = ("log_info", "log_error")


_logger = getLogger(__name__)
_MELONY_LOG_PREFIX: Final[str] = "[Melony]"
_DEBUG = True  # Toggle this for debugging via print python function  TODO: to settings
_WRAP_LINE_WIDTH: Final[int] = 60


def log_info(message: str, consumer_id: int | None = None) -> None:
    if _DEBUG:
        if consumer_id is None:
            print(f"{_MELONY_LOG_PREFIX}[INFO][{datetime.now()}]: {message}")  # noqa: WPS421, WPS226
        else:
            print(  # noqa: WPS421
                f"{_MELONY_LOG_PREFIX}[INFO][consumer-{consumer_id}][{datetime.now()}]: "  # noqa: WPS226
                f"{message}"
            )
    else:
        if consumer_id is None:
            _logger.info(f"{_MELONY_LOG_PREFIX}[INFO][{datetime.now()}]: {message}")
        else:
            _logger.info(
                f"{_MELONY_LOG_PREFIX}[INFO][consumer-{consumer_id}][{datetime.now()}]: {message}"
            )



def log_error(
    message: str,
    exc: BaseException | None = None,
    consumer_id: int | None = None
) -> None:
    if _DEBUG:
        print(f"{_MELONY_LOG_PREFIX}[ERROR][{datetime.now()}]: {message}")  # noqa: WPS421
        if exc:
            formatted_traceback = traceback.format_exception(exc)
            for line in formatted_traceback:
                print(line)  # noqa: WPS421
            print("-" * _WRAP_LINE_WIDTH)  # noqa: WPS421
    else:
        if consumer_id is None:
            _logger.error(
                f"{_MELONY_LOG_PREFIX}[ERROR][{datetime.now()}]: {message}", 
                exc_info=True
            )
        else:
            _logger.error(
                f"{_MELONY_LOG_PREFIX}[ERROR][consumer-{consumer_id}]"
                f"[{datetime.now()}]: {message}", 
                exc_info=True
            )



import json

from dataclasses import asdict, dataclass
from typing import Callable

__all__ = ()


@dataclass(frozen=True, kw_only=True, slots=True)
class CronTaskRegistration:
    func_path: str
    func: Callable
    cron: str
    queue: str
    retries: int
    retry_timeout: int
    is_coro: bool


@dataclass(frozen=True, kw_only=True, slots=True)
class CronEntry:
    func_path: str
    cron: str
    queue: str
    retries: int
    retry_timeout: int
    is_coro: bool

    def as_json(self) -> str:
        return json.dumps(asdict(self))

    @classmethod
    def from_registration(cls, reg: CronTaskRegistration) -> "CronEntry":
        return cls(
            func_path=reg.func_path,
            cron=reg.cron,
            queue=reg.queue,
            retries=reg.retries,
            retry_timeout=reg.retry_timeout,
            is_coro=reg.is_coro,
        )

    @classmethod
    def from_json(cls, serialized: str) -> "CronEntry":
        return cls(**json.loads(serialized))



from typing import Final

__all__ = ()

# default
QUEUE_PREFIX: Final[str] = "melony_tasks:"
CRON_QUEUE_PREFIX: Final[str] = "melony_cron:"
DEFAULT_QUEUE: Final[str] = f"default"

# redis
REDIS_RESULT_BACKEND_KEY: Final[str] = "melony_result_backend:"



import asyncio

from abc import ABC, abstractmethod
from functools import wraps
from inspect import signature
from typing import TYPE_CHECKING, Callable, ParamSpec, TypeVar, final, overload

from croniter import croniter

from melony.core.consts import CRON_QUEUE_PREFIX, DEFAULT_QUEUE
from melony.core.consumers import BaseConsumer
from melony.core.cron_tasks import CronTaskRegistration
from melony.core.publishers import Publisher
from melony.core.task_wrappers import AsyncTaskWrapper, SyncTaskWrapper, TaskWrapper

if TYPE_CHECKING:
    from melony.core.cron_consumers import BaseCronConsumer

__all__ = ()

_TaskParams = ParamSpec("_TaskParams")
_TaskResult = TypeVar("_TaskResult")


class BaseBroker(ABC):
    def __init__(self) -> None:
        self._cron_registry: list[CronTaskRegistration] = []

    @property
    @abstractmethod
    def publisher(self) -> Publisher:
        ...

    @property
    @abstractmethod
    def consumer(self) -> BaseConsumer:
        ...

    @property
    @abstractmethod
    def cron_consumer(self) -> "BaseCronConsumer":
        ...

    @overload
    def task(
        self,
        func: Callable[_TaskParams, _TaskResult]
    ) -> Callable[_TaskParams, TaskWrapper]:
        ...

    @overload
    def task(
        self,
        *,
        retries: int = 1,
        retry_timeout: int = 0,
        queue: str = DEFAULT_QUEUE,
        cron: str | None = None
    ) -> Callable[
            [Callable[_TaskParams, _TaskResult]],  # type: ignore
            Callable[_TaskParams, TaskWrapper]
        ]:
        ...

    @final
    def task(
        self,
        func: Callable[_TaskParams, _TaskResult] | None = None,
        *,
        retries: int = 1,
        retry_timeout: int = 0,
        queue: str = DEFAULT_QUEUE,
        cron: str | None = None
    ) -> Callable[_TaskParams, TaskWrapper] | Callable[[Callable], Callable]:
        """Decorator for registering your melony tasks.

        After your broker initizization (see selected broker documentation),
        you are able to register task for next delaying or execution.

        Task declaration example:

            >>> import asyncio

            >>> @broker.task
            >>> async def example_task(string_param: str) -> str:
            ...     asyncio.sleep(2)
            ...     return string_param.upper()

        Now you have opportunity to call special method for execute your task after some
        time. This method called .delay()

        Task delay example:
            >>> await example.task(string_param='my string').delay(countdown=30)  # will be executed after 30 seconds

        Or just execute this task as a common funcion immidiatly

        Task execution example:

            >>> await example_task(string_param='execute now').execute()
        """
        self._validate_params(retries, retry_timeout, cron)
        def _decorate(  # noqa: WPS430
            func: Callable[_TaskParams, _TaskResult]
        ) -> Callable[_TaskParams, TaskWrapper]:
            if cron:
                self._cron_registry.append(CronTaskRegistration(
                    func_path=f"{func.__module__}.{func.__qualname__}",
                    func=func,
                    cron=cron,
                    queue=f"{CRON_QUEUE_PREFIX}{queue}",
                    retries=retries,
                    retry_timeout=retry_timeout,
                    is_coro=asyncio.iscoroutinefunction(func),
                ))

            @wraps(func)
            def wrapper(
                *args: _TaskParams.args,
                **kwargs: _TaskParams.kwargs
            ) -> TaskWrapper:
                sig = signature(func)
                bound = sig.bind(*args, **kwargs)
                bound.apply_defaults()

                if asyncio.iscoroutinefunction(func):
                    wrapper.__annotations__ = {
                        **func.__annotations__,
                        "return": AsyncTaskWrapper
                    }
                    return AsyncTaskWrapper(
                        func=func,
                        broker=self,
                        bound_args=bound.arguments,
                        retries=retries,
                        retry_timeout=retry_timeout,
                        queue=queue
                    )
                else:
                    wrapper.__annotations__ = {
                        **func.__annotations__,
                        "return": SyncTaskWrapper
                    }
                    return SyncTaskWrapper(
                        func=func,
                        broker=self,
                        bound_args=bound.arguments,
                        retries=retries,
                        retry_timeout=retry_timeout,
                        queue=queue
                    )


            return wrapper

        if func is None:
            return _decorate
        else:
            return _decorate(func)

    @final
    def _validate_params(
        self,
        retries: int,
        retry_timeout: int,
        cron: str | None = None,
    ) -> None:
        if retries <= 0:
            raise ValueError(
                "Parameter 'retries' must be positive integer (without zero)"
            )
        if retry_timeout < 0:
            raise ValueError(
                "Parameter 'retry_timeout' must be positive integer or zero"
            )
        if cron is not None and not croniter.is_valid(cron):
            raise ValueError(f"Invalid cron expression: '{cron}'")



import asyncio
import multiprocessing
import time

from abc import ABC, abstractmethod
from typing import Awaitable, Iterable, final
from redis.exceptions import ConnectionError

from melony.core.consts import DEFAULT_QUEUE, QUEUE_PREFIX
from melony.core.dto import FilteredTasksDTO, TaskExecResultsDTO
from melony.core.publishers import IAsyncPublisher, ISyncPublisher
from melony.core.result_backends import (
    IAsyncResultBackendSaver,
    IResultBackend,
    ISyncResultBackendSaver
)
from melony.core.task_executor import AsyncTaskExecutor, SyncTaskExecutor
from melony.core.tasks import Task
from melony.logger import log_error, log_info


__all__ = ()


class BaseConsumer:
    @final
    def _run_all_processes(self, processes: list) -> None:
        for process in processes:
            process.start()
        for process in processes:
            process.join()

    @final
    def _filter_tasks_by_execution_time(
        self,
        tasks: Iterable[Task],
        consumer_id
    ) -> FilteredTasksDTO:
        tasks_to_execute: list[Task] = []
        tasks_to_push_back: list[Task] = []
        current_timestamp = time.time()
        
        for task in tasks:
            task_execution_timestamp = task.get_execution_timestamp()
            if task_execution_timestamp < current_timestamp:
                tasks_to_execute.append(task)
            else:
                tasks_to_push_back.append(task)
        
        if tasks_to_execute:
            log_info(f"Executing tasks: {tasks_to_execute}", consumer_id=consumer_id)
        return FilteredTasksDTO(
            tasks_to_execute=tasks_to_execute,
            tasks_to_push_back=tasks_to_push_back
        )


class BaseAsyncConsumer(ABC, BaseConsumer):  # noqa: WPS214
    def __init__(
        self,
        publisher: IAsyncPublisher,
        result_backend: IResultBackend | None = None
    ) -> None:
        self._publisher = publisher
        self._connection = self._publisher.connection
        self._result_backend = result_backend

        result_backend_saver = result_backend.saver if result_backend else None
        if result_backend_saver:
            assert isinstance(result_backend_saver, IAsyncResultBackendSaver)
        
        self._task_executor = AsyncTaskExecutor(
            connection=self._connection,
            result_backend_saver=result_backend_saver
        )

    @final
    async def start_consume(
        self,
        queue: str = DEFAULT_QUEUE,
        processes: int = 1
    ) -> None:
        queue = f"{QUEUE_PREFIX}{queue}"

        if processes < 1:
            raise ValueError("Param 'processes' must be positive integer (without zero)")

        if processes == 1:
            await self._consumer_loop(queue=queue, consumer_id=0)
            return

        running_processes = []
        for process_num in range(processes):
            process = multiprocessing.Process(
                name=f"melony-process-{process_num}",
                target=self._run_consumer_in_process,
                args=(queue, process_num),
                daemon=False,
            )
            running_processes.append(process)
        try:
            self._run_all_processes(running_processes)
        except KeyboardInterrupt:
            for proc in running_processes:
                if proc.is_alive():
                    proc.terminate()

    @abstractmethod
    async def _pop_tasks(self, queue: str) -> Iterable[Task]:
        """Get all tasks from message broker."""

    @final
    async def _consumer_loop(self, queue: str, consumer_id: int = 1) -> None:
        log_info(f"Start listening queue {queue}", consumer_id=consumer_id)
        while True:
            try:
                await self._consumer_loop_iteration(queue, consumer_id)
            except ConnectionError as exc:
                log_error(f"Redis connection error", exc=exc)
                break
            except Exception as exc:
                log_error(f"Unexpected error at consuming loop", exc=exc)
                break

    @final
    async def _consumer_loop_iteration(self, queue: str, consumer_id: int) -> None:
        tasks = await self._pop_tasks(queue=queue)
        filtered_tasks = self._filter_tasks_by_execution_time(tasks, consumer_id)
        await self._push_bulk(tasks=filtered_tasks.tasks_to_push_back)
        wait_task_results = await self._task_executor.execute_tasks(
            tasks=filtered_tasks.tasks_to_execute,
            consumer_id=consumer_id
        )
        await self._retry_policy(wait_task_results)

    @final
    async def _push_bulk(self, tasks: Iterable[Task]) -> None:
        push_coroutines = [self._publisher.push(task) for task in tasks]
        await asyncio.gather(*[coro for coro in push_coroutines if coro is not None])

    @final
    async def _retry_policy(
        self,
        tasks_exec_results: TaskExecResultsDTO
    ) -> None:
        tasks_to_push_back: list[Task] = []
        for task_to_retry in tasks_exec_results.tasks_to_retry:
            if task_to_retry._meta.retries_left:
                if task_to_retry.retry_timeout:
                    task_to_retry._meta.timestamp += task_to_retry.retry_timeout
                tasks_to_push_back.append(task_to_retry)

        await self._push_bulk(tasks_to_push_back)

    @final
    def _run_consumer_in_process(
        self,
        queue: str,
        consumer_id: int = 0
    ) -> None:
        asyncio.run(self._consumer_loop(queue=queue, consumer_id=consumer_id))


class BaseSyncConsumer(ABC, BaseConsumer):
    def __init__(
        self,
        publisher: ISyncPublisher,
        result_backend: IResultBackend | None = None
    ) -> None:
        self._publisher = publisher
        self._connection = self._publisher.connection
        self._result_backend = result_backend

        result_backend_saver = result_backend.saver if result_backend else None
        if result_backend_saver:
            assert isinstance(result_backend_saver, ISyncResultBackendSaver)

        self._task_executor = SyncTaskExecutor(
            connection=self._connection,
            result_backend_saver=result_backend_saver
        )

    # TODO: fix that typing
    @final
    def start_consume(self, queue: str = DEFAULT_QUEUE, processes: int = 1) -> Awaitable:  # type: ignore
        queue = f"{QUEUE_PREFIX}{queue}"

        if processes < 1:
            raise ValueError("Param 'processes' must be positive integer (without zero)")

        if processes == 1:
            self._consumer_loop(queue=queue, consumer_id=0)
            return  # type: ignore

        for process_num in range(processes):
            process = multiprocessing.Process(
                name=f"melony-process-{process_num}",
                target=self._consumer_loop,
                args=(queue, process_num),
                daemon=False
            )
            process.start()

    @abstractmethod
    def _pop_tasks(self, queue: str) -> Iterable[Task]:
        """Get all tasks from message broker."""

    @final
    def _consumer_loop(self, queue: str, consumer_id: int = 1) -> None:
        log_info(f"Start listening queue {queue}", consumer_id=consumer_id)
        while True:
            try:
                self._consumer_loop_iteration(queue, consumer_id)
            except ConnectionError as exc:
                log_error(f"Redis connection error", exc=exc)
                break
            except Exception as exc:
                log_error(f"Unexpected error at consuming loop", exc=exc)
                break

    @final
    def _consumer_loop_iteration(self, queue: str, consumer_id: int) -> None:
        tasks = self._pop_tasks(queue)
        filtered_tasks = self._filter_tasks_by_execution_time(tasks, consumer_id)
        self._push_bulk(tasks=filtered_tasks.tasks_to_push_back)
        wait_task_results = self._task_executor.execute_tasks(
            tasks=filtered_tasks.tasks_to_execute,
            consumer_id=consumer_id
        )
        self._retry_policy(wait_task_results)

    @final
    def _push_bulk(self, tasks: Iterable[Task]) -> None:
        for task in tasks:
            self._publisher.push(task)

    @final
    def _retry_policy(
        self,
        tasks_exec_results: TaskExecResultsDTO
    ) -> None:
        tasks_to_push_back: list[Task] = []
        for task_to_retry in tasks_exec_results.tasks_to_retry:
            if task_to_retry._meta.retries_left:
                if task_to_retry.retry_timeout:
                    task_to_retry._meta.timestamp += task_to_retry.retry_timeout
                tasks_to_push_back.append(task_to_retry)

        self._push_bulk(tasks_to_push_back)



from abc import ABC, abstractmethod
from redis.asyncio import Redis
from redis import Redis as SyncRedis

from melony.core.tasks import Task


__all__ = ()


type Publisher = IAsyncPublisher | ISyncPublisher


class IAsyncPublisher(ABC):
    @property
    @abstractmethod
    def connection(self) -> Redis:
        """Aync redis connection getter."""

    @abstractmethod
    async def push(self, task: Task) -> None:
        """Async push your melony task to message broker."""

class ISyncPublisher(ABC):
    @property
    @abstractmethod
    def connection(self) -> SyncRedis:
        """Sync redis connection getter."""

    @abstractmethod
    def push(self, task: Task) -> None:
        """Sync push your melony task to message broker."""



import asyncio
import json

from typing import Iterable, Sequence, final, override
from redis.asyncio import Redis
from redis import Redis as SyncRedis

from melony.core.consts import REDIS_RESULT_BACKEND_KEY
from melony.core.dto import TaskResultDTO
from melony.core.result_backends import (
    IAsyncResultBackendSaver,
    IResultBackend,
    ISyncResultBackendSaver,
    ResultBackendSaver
)


__all__ = ()


@final
class _AsyncRedisResultBackendSaver(IAsyncResultBackendSaver):
    def __init__(self, redis_connection: Redis | SyncRedis) -> None:
        self._connection = redis_connection

    @override
    async def save_results(self, task_results: Iterable[TaskResultDTO]) -> None:
        set_result_coroutines = []

        for task_result_info in task_results:
            task_id = task_result_info.task.task_id
            redis_key = f"{REDIS_RESULT_BACKEND_KEY}{task_id}"
            save_task_data = {}
            save_task_data["result"] = task_result_info.task_result
            save_task_data["task"] = task_result_info.task.as_dict()
            set_result_coroutines.append(
                self._connection.set(
                    name=redis_key,
                    value=json.dumps(save_task_data)
                )
            )
        await asyncio.gather(*set_result_coroutines)


@final
class _SyncRedisResultBackendSaver(ISyncResultBackendSaver):
    def __init__(self, redis_connection: Redis | SyncRedis) -> None:
        self._connection = redis_connection

    @override
    def save_results(self, task_results: Iterable[TaskResultDTO]) -> None:
        for task_result_info in task_results:
            task_id = task_result_info.task.task_id
            redis_key = f"{REDIS_RESULT_BACKEND_KEY}{task_id}"
            save_task_data = {}
            save_task_data["result"] = task_result_info.task_result
            save_task_data["task"] = task_result_info.task.as_dict()
            self._connection.set(
                name=redis_key,
                value=json.dumps(save_task_data)
            )

@final
class RedisResultBackend(IResultBackend):
    def __init__(self, redis_connection: Redis | SyncRedis) -> None:
        self._connection = redis_connection
    
    @property
    def saver(self) -> ResultBackendSaver:
        match self._connection:
            case Redis():
                return _AsyncRedisResultBackendSaver(redis_connection=self._connection)
            case SyncRedis():
                return _SyncRedisResultBackendSaver(redis_connection=self._connection)
            case _:
                assert_never(_)




from typing import assert_never, final, override
from redis.asyncio.client import Redis
from redis import Redis as SyncRedis

from melony.core.publishers import IAsyncPublisher, ISyncPublisher, Publisher
from melony.core.brokers import BaseBroker
from melony.core.result_backends import (
    IAsyncResultBackendSaver,
    IResultBackend,
    ISyncResultBackendSaver,
)
from melony.redis_broker.consumers import (
    AsyncRedisConsumer,
    SyncRedisConsumer,
    RedisConsumer
)
from melony.redis_broker.cron_consumers import (
    AsyncRedisCronConsumer,
    RedisCronConsumer,
    SyncRedisCronConsumer,
)
from melony.redis_broker.publishers import AsyncRedisPublisher, SyncRedisPublisher


__all__ = ()


@final
class RedisBroker(BaseBroker):
    """Melony redis broker implementation.

    Needs for task registration.
    Maybe async or sync. It depends from type of your provided connection.

    Async initialization example:

        >>> from melony import RedisBroker
        >>> from redis.asyncio import Redis

        >>> async_connection = Redis(host='localhost', port=6379, db=0)
        >>> broker = RedisBroker(redis_connection=async_connection)

    Sync initialization example:

        >>> from melony import RedisBroker
        >>> from redis import Redis as SyncRedis

        >>> sync_connection = SyncRedis(host='localhost', port=6379, db=0)
        >>> broker=RedisBroker(redis_connection=sync_connection)

    You are able to provide result backend as well.

    Result backend example:

        >>> from melony import RedisBroker, RedisResultBackend  # or another result backend
        >>> from redis.asyncio import Redis

        >>> connection = Redis(host='localhost', port=6379, db=0)
        >>> result_backend = RedisResultBackend()
        >>> broker = RedisBroker(redis_connection=connection, result_backend=result_backend)

    See .task() method documentation for registering your tasks.

    Also you have opportunity to call special core properties: publisher and consumer.
    """
    def __init__(
        self,
        redis_connection: Redis | SyncRedis,
        result_backend: IResultBackend | None = None
    ) -> None:
        super().__init__()
        self._connection = redis_connection
        self._result_backend = result_backend

    @property
    @override
    def publisher(self) -> Publisher:
        """Getter for melony publisher for redis broker.

        For advanced users only. Usually you have not to use it.

        But if you want, you are able to push your tasks handly by .push() method of your
        publisher.

        For doing this, you should serialize task first, you are able to use your own
        serializer or converter which already exists in melony.core.task_converters
        """
        match self._connection:
            case Redis():
                return AsyncRedisPublisher(connection=self._connection)
            case SyncRedis():
                return SyncRedisPublisher(connection=self._connection)
            case _:
                assert_never(_)

    @property
    @override
    def consumer(self) -> RedisConsumer:
        """Melony redis consumer implementation.

        Needs for listening messages from redis broker, executing your tasks and writing
        result backend to selected ResultBackend (see more about result backend at
        documentation).

        For start consuming your tasks you need to call .start_consume() method.

        Run consumer example:

            >>> @broker.task
            >>> async def example_task(string_param: str) -> str:
            >>>     asyncio.sleep(2)
            >>>     return string_param.upper()

            >>> broker.consumer.start_consume()  # Run consuming process.

        Also you are able to provide 'processes' parameter for choosing number of processes
        (workers).

        Providing processes param example:

            >>> broker.consumer.start_consume(processes=3)

        There is no recomendation how many procceses you should use. It depends from many
        things. So, just try to understand optimal value imperatively.
        """
        match self._connection:
            case Redis():
                assert isinstance(self.publisher, IAsyncPublisher)
                if self._result_backend:
                    assert isinstance(
                        self._result_backend.saver,
                        IAsyncResultBackendSaver
                    )
                return AsyncRedisConsumer(
                    publisher=self.publisher,
                    broker=self,
                    result_backend=self._result_backend
                )
            case SyncRedis():
                assert isinstance(self.publisher, ISyncPublisher)
                if self._result_backend:
                    assert isinstance(
                        self._result_backend.saver,
                        ISyncResultBackendSaver
                    )
                return SyncRedisConsumer(
                    publisher=self.publisher,
                    broker=self,
                    result_backend=self._result_backend
                )
            case _:
                assert_never(self._connection)

    @property
    @override
    def cron_consumer(self) -> RedisCronConsumer:
        """Melony redis cron consumer implementation.

        Needs for executing scheduled (cron) tasks registered with the cron parameter.

        Run cron consumer example:

            >>> @broker.task(cron="* * * * *")
            >>> async def scheduled_task() -> None:
            >>>     ...

            >>> await broker.cron_consumer.start_consume()
        """
        match self._connection:
            case Redis():
                return AsyncRedisCronConsumer(
                    connection=self._connection,
                    broker=self,
                    result_backend=self._result_backend,
                )
            case SyncRedis():
                return SyncRedisCronConsumer(
                    connection=self._connection,
                    broker=self,
                    result_backend=self._result_backend,
                )
            case _:
                assert_never(self._connection)



import json

from datetime import datetime
from dataclasses import asdict, dataclass, field
from inspect import unwrap
from typing import (
    Callable,
    Any,
    TYPE_CHECKING,
    final,
    Final
)


if TYPE_CHECKING:
    from melony.core.brokers import BaseBroker

__all__ = (
    "AsyncTask",
    "SyncTask"
)

type Task = AsyncTask | SyncTask


_MAX_COUNTDOWN_SEC: Final[int] = 86400
_MAX_COUNTDOWN_MIN: Final[float] = _MAX_COUNTDOWN_SEC / 60


@final
@dataclass(kw_only=True, slots=True)
class _TaskMeta:
    retries_left: int | None = None
    timestamp: float = field(default_factory=lambda: datetime.timestamp(datetime.now()))


@dataclass(frozen=True, kw_only=True, slots=True)
class _BaseTask:
    task_id: str
    kwargs: dict[str, Any]
    countdown: int
    retries: int
    retry_timeout: int
    queue: str
    is_coro: bool

    _meta: _TaskMeta = field(default_factory=_TaskMeta)

    def __post_init__(self) -> None:
        self._validate_countdown()
        self._set_retries_left_to_meta()

    @final
    @property
    def timestamp(self) -> float:
        return self._meta.timestamp

    @final
    def get_execution_timestamp(self) -> float:
        return self.timestamp + self.countdown

    @final
    def _validate_countdown(self) -> None:
        if self.countdown < 0:
            raise ValueError("Countdown cannot be negative")
        elif self.countdown >= _MAX_COUNTDOWN_SEC:
            raise ValueError(
                f"Countdown cannot be greater than {_MAX_COUNTDOWN_SEC} "
                f"({_MAX_COUNTDOWN_MIN} minutes)"
            )

    @final
    def _set_retries_left_to_meta(self) -> None:
        if not self._meta.retries_left:
            self._meta.retries_left = self.retries


@final
@dataclass(frozen=True, kw_only=True, slots=True)
class _TaskJSONSerializable(_BaseTask):
    func_name: str
    func_path: str

@dataclass(frozen=True, kw_only=True, slots=True)
class _Task(_BaseTask):
    func: Callable
    func_path: str
    broker: "BaseBroker"

    @final
    def as_json_serializable_obj(self) -> _TaskJSONSerializable:
        return _TaskJSONSerializable(
            task_id=self.task_id,
            kwargs=self.kwargs,
            countdown=self.countdown,
            func_name=self.func.__name__,
            func_path=self.func_path,
            retries=self.retries,
            retry_timeout=self.retry_timeout,
            queue=self.queue,
            is_coro=self.is_coro,
            _meta=self._meta,
        )

    @final
    def as_dict(self) -> dict[str, Any]:
        return asdict(self.as_json_serializable_obj())

    @final
    def as_json(self) -> str:
        json_str = json.dumps(self.as_dict())
        return json_str


@final
@dataclass(frozen=True, kw_only=True, slots=True)
class AsyncTask(_Task):
    async def execute(self) -> Any:
        unwrapped_func = unwrap(self.func)
        task_result = await unwrapped_func(**self.kwargs)
        return task_result

@final
@dataclass(frozen=True, kw_only=True, slots=True)
class SyncTask(_Task):
    def execute(self) -> Any:
        unwrapped_func = unwrap(self.func)
        task_result = unwrapped_func(**self.kwargs)
        return task_result


import pytest

from unittest.mock import MagicMock, patch
from redis.exceptions import ConnectionError

from melony.core.cron_consumers import BaseAsyncCronConsumer, BaseSyncCronConsumer
from melony.core.cron_tasks import CronEntry


_QUEUE = "melony_cron:default"


def _make_entry(**kwargs) -> CronEntry:
    defaults = dict(
        func_path="mymod.my_func",
        cron="* * * * *",
        queue=_QUEUE,
        retries=3,
        retry_timeout=0,
        is_coro=False,
    )
    return CronEntry(**{**defaults, **kwargs})


class _AsyncStub(BaseAsyncCronConsumer):
    def __init__(self, broker=None, due_batches=None):
        super().__init__(connection=None, broker=broker or MagicMock())
        self._due_batches: list[list[CronEntry]] = due_batches or []
        self.rescheduled: list[CronEntry] = []
        self.initialized: list[str] = []

    async def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
        if self._due_batches:
            return self._due_batches.pop(0)
        return []

    async def _initialize_cron_tasks(self, cron_queue: str) -> None:
        self.initialized.append(cron_queue)

    async def _reschedule(self, entry: CronEntry, cron_queue: str) -> None:
        self.rescheduled.append(entry)


class _SyncStub(BaseSyncCronConsumer):
    def __init__(self, broker=None, due_batches=None):
        super().__init__(connection=None, broker=broker or MagicMock())
        self._due_batches: list[list[CronEntry]] = due_batches or []
        self.rescheduled: list[CronEntry] = []
        self.initialized: list[str] = []

    def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
        if self._due_batches:
            return self._due_batches.pop(0)
        return []

    def _initialize_cron_tasks(self, cron_queue: str) -> None:
        self.initialized.append(cron_queue)

    def _reschedule(self, entry: CronEntry, cron_queue: str) -> None:
        self.rescheduled.append(entry)


async def test_async_start_consume_rejects_zero_processes():
    consumer = _AsyncStub()
    with pytest.raises(ValueError, match="processes"):
        await consumer.start_consume(processes=0)


async def test_async_start_consume_rejects_negative_processes():
    consumer = _AsyncStub()
    with pytest.raises(ValueError):
        await consumer.start_consume(processes=-5)


def test_sync_start_consume_rejects_zero_processes():
    consumer = _SyncStub()
    with pytest.raises(ValueError, match="processes"):
        consumer.start_consume(processes=0)


async def test_async_execute_entry_success_first_attempt():
    call_count = 0

    async def my_func():
        nonlocal call_count
        call_count += 1

    entry = _make_entry(is_coro=True)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=my_func):
        await consumer._execute_entry(entry, _QUEUE, 0)

    assert call_count == 1
    assert consumer.rescheduled == [entry]


async def test_async_execute_entry_retries_until_success():
    call_count = 0

    async def flaky():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise RuntimeError("temporary failure")

    entry = _make_entry(is_coro=True, retries=3)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=flaky):
        await consumer._execute_entry(entry, _QUEUE, 0)

    assert call_count == 3
    assert len(consumer.rescheduled) == 1


async def test_async_execute_entry_reschedules_after_all_retries_exhausted():
    async def always_fails():
        raise ValueError("boom")

    entry = _make_entry(is_coro=True, retries=3)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        await consumer._execute_entry(entry, _QUEUE, 0)

    assert len(consumer.rescheduled) == 1


async def test_async_execute_entry_exhausts_all_retries():
    call_count = 0

    async def always_fails():
        nonlocal call_count
        call_count += 1
        raise ValueError("boom")

    entry = _make_entry(is_coro=True, retries=5)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        await consumer._execute_entry(entry, _QUEUE, 0)

    assert call_count == 5


async def test_async_execute_entry_calls_sync_func_directly():
    call_count = 0

    def sync_func():
        nonlocal call_count
        call_count += 1

    entry = _make_entry(is_coro=False)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=sync_func):
        await consumer._execute_entry(entry, _QUEUE, 0)

    assert call_count == 1


async def test_async_execute_entry_sleeps_between_failed_attempts():
    async def always_fails():
        raise RuntimeError()

    entry = _make_entry(is_coro=True, retries=3, retry_timeout=5)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        with patch("asyncio.sleep") as mock_sleep:
            await consumer._execute_entry(entry, _QUEUE, 0)

    assert mock_sleep.call_count == 2
    mock_sleep.assert_called_with(5)


async def test_async_execute_entry_no_sleep_after_last_attempt():
    async def always_fails():
        raise RuntimeError()

    entry = _make_entry(is_coro=True, retries=1, retry_timeout=10)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        with patch("asyncio.sleep") as mock_sleep:
            await consumer._execute_entry(entry, _QUEUE, 0)

    assert mock_sleep.call_count == 0


async def test_async_execute_entry_no_sleep_on_success():
    async def my_func():
        pass

    entry = _make_entry(is_coro=True, retries=3, retry_timeout=10)
    consumer = _AsyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=my_func):
        with patch("asyncio.sleep") as mock_sleep:
            await consumer._execute_entry(entry, _QUEUE, 0)

    assert mock_sleep.call_count == 0


def test_sync_execute_entry_success_first_attempt():
    call_count = 0

    def my_func():
        nonlocal call_count
        call_count += 1

    entry = _make_entry(is_coro=False)
    consumer = _SyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=my_func):
        consumer._execute_entry(entry, _QUEUE, 0)

    assert call_count == 1
    assert consumer.rescheduled == [entry]


def test_sync_execute_entry_retries_until_success():
    call_count = 0

    def flaky():
        nonlocal call_count
        call_count += 1
        if call_count < 2:
            raise RuntimeError("fail once")

    entry = _make_entry(is_coro=False, retries=3)
    consumer = _SyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=flaky):
        consumer._execute_entry(entry, _QUEUE, 0)

    assert call_count == 2
    assert len(consumer.rescheduled) == 1


def test_sync_execute_entry_reschedules_after_all_retries_exhausted():
    def always_fails():
        raise ValueError()

    entry = _make_entry(is_coro=False, retries=3)
    consumer = _SyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        consumer._execute_entry(entry, _QUEUE, 0)

    assert len(consumer.rescheduled) == 1


def test_sync_execute_entry_exhausts_all_retries():
    call_count = 0

    def always_fails():
        nonlocal call_count
        call_count += 1
        raise ValueError()

    entry = _make_entry(is_coro=False, retries=4)
    consumer = _SyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        consumer._execute_entry(entry, _QUEUE, 0)

    assert call_count == 4


def test_sync_execute_entry_sleeps_between_failed_attempts():
    def always_fails():
        raise RuntimeError()

    entry = _make_entry(is_coro=False, retries=3, retry_timeout=10)
    consumer = _SyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        with patch("time.sleep") as mock_sleep:
            consumer._execute_entry(entry, _QUEUE, 0)

    assert mock_sleep.call_count == 2
    mock_sleep.assert_called_with(10)


def test_sync_execute_entry_no_sleep_after_last_attempt():
    def always_fails():
        raise RuntimeError()

    entry = _make_entry(is_coro=False, retries=1, retry_timeout=10)
    consumer = _SyncStub()

    with patch("melony.core.cron_consumers.find_task_func", return_value=always_fails):
        with patch("time.sleep") as mock_sleep:
            consumer._execute_entry(entry, _QUEUE, 0)

    assert mock_sleep.call_count == 0


async def test_async_loop_iteration_executes_all_due_entries():
    entry1 = _make_entry(func_path="m.f1", is_coro=True)
    entry2 = _make_entry(func_path="m.f2", is_coro=True)

    async def noop():
        pass

    consumer = _AsyncStub(due_batches=[[entry1, entry2]])

    with patch("melony.core.cron_consumers.find_task_func", return_value=noop):
        await consumer._consumer_loop_iteration(_QUEUE, 0)

    assert consumer.rescheduled == [entry1, entry2]


async def test_async_loop_iteration_empty_batch_no_reschedule():
    consumer = _AsyncStub(due_batches=[[]])
    await consumer._consumer_loop_iteration(_QUEUE, 0)
    assert consumer.rescheduled == []


def test_sync_loop_iteration_executes_all_due_entries():
    entry1 = _make_entry(func_path="m.f1")
    entry2 = _make_entry(func_path="m.f2")

    def noop():
        pass

    consumer = _SyncStub(due_batches=[[entry1, entry2]])

    with patch("melony.core.cron_consumers.find_task_func", return_value=noop):
        consumer._consumer_loop_iteration(_QUEUE, 0)

    assert consumer.rescheduled == [entry1, entry2]


async def test_async_consumer_loop_breaks_on_connection_error():
    class _ErrorStub(_AsyncStub):
        async def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
            raise ConnectionError("Redis down")

    consumer = _ErrorStub()
    await consumer._consumer_loop(_QUEUE, 0)


async def test_async_consumer_loop_breaks_on_unexpected_exception():
    class _ErrorStub(_AsyncStub):
        async def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
            raise RuntimeError("unexpected")

    consumer = _ErrorStub()
    await consumer._consumer_loop(_QUEUE, 0)


async def test_async_consumer_loop_initializes_before_iterating():
    class _ErrorStub(_AsyncStub):
        async def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
            raise ConnectionError()

    consumer = _ErrorStub()
    await consumer._consumer_loop("melony_cron:reports", 0)

    assert "melony_cron:reports" in consumer.initialized


def test_sync_consumer_loop_breaks_on_connection_error():
    class _ErrorStub(_SyncStub):
        def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
            raise ConnectionError("Redis down")

    consumer = _ErrorStub()
    consumer._consumer_loop(_QUEUE, 0)


def test_sync_consumer_loop_breaks_on_unexpected_exception():
    class _ErrorStub(_SyncStub):
        def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
            raise RuntimeError("unexpected")

    consumer = _ErrorStub()
    consumer._consumer_loop(_QUEUE, 0)


def test_sync_consumer_loop_initializes_before_iterating():
    class _ErrorStub(_SyncStub):
        def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
            raise ConnectionError()

    consumer = _ErrorStub()
    consumer._consumer_loop("melony_cron:emails", 0)

    assert "melony_cron:emails" in consumer.initialized



from collections.abc import Awaitable, Callable
from inspect import signature
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar, final
from uuid import uuid4

from melony.core.consts import QUEUE_PREFIX
from melony.core.tasks import AsyncTask, SyncTask
from melony.logger import log_info

if TYPE_CHECKING:
    from melony.core.brokers import BaseBroker

__all__ = ()


type TaskWrapper = AsyncTaskWrapper | SyncTaskWrapper

_TaskParams = ParamSpec("_TaskParams")
_TaskResult = TypeVar("_TaskResult")


class _BaseTaskWrapper(Awaitable):
    @final
    def __init__(  # noqa: WPS211
        self,
        func: Callable[_TaskParams, _TaskResult],
        broker: "BaseBroker",
        bound_args: dict[str, Any],
        retries: int,
        retry_timeout: int,
        queue: str
    ) -> None:
        self._func = func
        self._broker = broker
        self._sig = signature(func)
        self._bound_args = bound_args or {}
        self._func_path = f"{func.__module__}.{func.__qualname__}"
        self._retries = retries
        self._retry_timeout = retry_timeout
        self._queue = f"{QUEUE_PREFIX}{queue}"

    @final
    def __call__(
        self,
        *args: _TaskParams.args,  # type: ignore
        **kwargs: _TaskParams.kwargs
    ) -> "_BaseTaskWrapper":
        bound = self._sig.bind(*args, **kwargs)
        bound.apply_defaults()
        self._bound_args = bound.arguments
        return self

    @final
    def __await__(self) -> Any:
        raise RuntimeError(
            "Task cannot be awaited, did you forget to call .delay()?"
        )

@final
class AsyncTaskWrapper(_BaseTaskWrapper):
    async def delay(self, countdown: int = 0) -> AsyncTask:
        from melony.core.publishers import IAsyncPublisher
        assert isinstance(self._broker.publisher, IAsyncPublisher)
        task = AsyncTask(
            task_id=str(uuid4()),
            func=self._func,
            func_path=self._func_path,
            kwargs=self._bound_args,
            broker=self._broker,
            countdown=countdown,
            retries=self._retries,
            retry_timeout=self._retry_timeout,
            queue=self._queue,
            is_coro=True
        )
        await self._broker.publisher.push(task)
        log_info(f"Pushed task {task} to queue {task.queue}")
        return task


@final
class SyncTaskWrapper(_BaseTaskWrapper):
    def delay(self, countdown: int = 0) -> Awaitable[SyncTask]:  # TODO: need to fix typing here
        from melony.core.publishers import ISyncPublisher
        assert isinstance(self._broker.publisher, ISyncPublisher)
        task = SyncTask(
            task_id=str(uuid4()),
            func=self._func,
            func_path=self._func_path,
            kwargs=self._bound_args,
            broker=self._broker,
            countdown=countdown,
            retries=self._retries,
            retry_timeout=self._retry_timeout,
            queue=self._queue,
            is_coro=False
        )
        self._broker.publisher.push(task)
        log_info(f"Pushed task {task} to queue {task.queue}")
        return task  # type: ignore



import json

import pytest

from melony.core.cron_tasks import CronEntry, CronTaskRegistration


def _make_entry(**kwargs) -> CronEntry:
    defaults = dict(
        func_path="mymod.my_func",
        cron="* * * * *",
        queue="melony_cron:default",
        retries=3,
        retry_timeout=0,
        is_coro=False,
    )
    return CronEntry(**{**defaults, **kwargs})


def test_as_json_produces_valid_json():
    entry = _make_entry()
    data = json.loads(entry.as_json())
    assert data["func_path"] == "mymod.my_func"
    assert data["cron"] == "* * * * *"
    assert data["queue"] == "melony_cron:default"
    assert data["retries"] == 3
    assert data["retry_timeout"] == 0
    assert data["is_coro"] is False


def test_from_json_round_trip():
    entry = _make_entry(is_coro=True, retry_timeout=30, retries=5)
    assert CronEntry.from_json(entry.as_json()) == entry


def test_from_json_restores_all_fields():
    entry = _make_entry(func_path="tasks.send_email", cron="0 9 * * 1", retries=2)
    restored = CronEntry.from_json(entry.as_json())
    assert restored.func_path == entry.func_path
    assert restored.cron == entry.cron
    assert restored.queue == entry.queue
    assert restored.retries == entry.retries
    assert restored.retry_timeout == entry.retry_timeout
    assert restored.is_coro == entry.is_coro


def test_from_registration_maps_all_fields():
    def my_func():
        pass

    reg = CronTaskRegistration(
        func_path="tasks.my_func",
        func=my_func,
        cron="0 12 * * *",
        queue="melony_cron:reports",
        retries=2,
        retry_timeout=10,
        is_coro=False,
    )
    entry = CronEntry.from_registration(reg)

    assert entry.func_path == reg.func_path
    assert entry.cron == reg.cron
    assert entry.queue == reg.queue
    assert entry.retries == reg.retries
    assert entry.retry_timeout == reg.retry_timeout
    assert entry.is_coro == reg.is_coro


def test_from_registration_does_not_include_func():
    def my_func():
        pass

    reg = CronTaskRegistration(
        func_path="m.f", func=my_func, cron="* * * * *",
        queue="melony_cron:default", retries=1, retry_timeout=0, is_coro=False,
    )
    entry = CronEntry.from_registration(reg)
    assert not hasattr(entry, "func")


def test_from_registration_is_coro_true():
    async def my_coro():
        pass

    reg = CronTaskRegistration(
        func_path="m.my_coro", func=my_coro, cron="* * * * *",
        queue="melony_cron:default", retries=1, retry_timeout=0, is_coro=True,
    )
    entry = CronEntry.from_registration(reg)
    assert entry.is_coro is True



from dataclasses import dataclass
from typing import Any, Iterable

from melony.core.tasks import Task


__all__ = ()


@dataclass(frozen=True, kw_only=True, slots=True)
class FilteredTasksDTO:
    tasks_to_execute: Iterable[Task]
    tasks_to_push_back: Iterable[Task]


@dataclass(frozen=True, kw_only=True, slots=True)
class TaskResultDTO:
    task: Task
    task_result: Any


@dataclass(frozen=True, kw_only=True, slots=True)
class TaskExecResultsDTO:
    tasks_with_result: Iterable[TaskResultDTO]
    tasks_to_retry: Iterable[Task]


@dataclass(frozen=True, kw_only=True, slots=True)
class QueueMetaDTO:
    queue: str
    is_consumed: bool
    



import json

from abc import ABC, abstractmethod

from melony.core.brokers import BaseBroker
from melony.core.task_finders import find_task_func
from melony.core.tasks import _TaskMeta, AsyncTask, SyncTask, Task


__all__ = ("TaskConverter",)


class _BaseTaskConverter(ABC):
    @abstractmethod
    def deserialize_task(
        self,
        serialized_task: str,
        broker: BaseBroker
    ) -> Task:
        ...

    def serialize_task(self, task: Task) -> str:
        return task.as_json()


class TaskConverter(_BaseTaskConverter):
    def deserialize_task(
        self,
        serialized_task: str,
        broker: BaseBroker
    ) -> Task:
        task_dict = json.loads(serialized_task)
        task_func_path = task_dict["func_path"]
        task_func = find_task_func(task_func_path)

        if task_dict["is_coro"]:
            return AsyncTask(
                task_id=task_dict["task_id"],
                kwargs=task_dict["kwargs"],
                countdown=task_dict["countdown"],
                func=task_func,
                func_path=task_func_path,
                broker=broker,
                retries=task_dict["retries"],
                retry_timeout=task_dict["retry_timeout"],
                queue=task_dict["queue"],
                is_coro=task_dict["is_coro"],
                _meta=_TaskMeta(**task_dict["_meta"])
            )
        else:
            return SyncTask(
                task_id=task_dict["task_id"],
                kwargs=task_dict["kwargs"],
                countdown=task_dict["countdown"],
                func=task_func,
                func_path=task_func_path,
                broker=broker,
                retries=task_dict["retries"],
                retry_timeout=task_dict["retry_timeout"],
                queue=task_dict["queue"],
                is_coro=task_dict["is_coro"],
                _meta=_TaskMeta(**task_dict["_meta"])
            )


from datetime import datetime
from typing import TYPE_CHECKING, final, override

from croniter import croniter
from redis.asyncio.client import Redis
from redis import Redis as SyncRedis

from melony.core.cron_consumers import BaseAsyncCronConsumer, BaseSyncCronConsumer
from melony.core.cron_tasks import CronEntry

if TYPE_CHECKING:
    from melony.core.brokers import BaseBroker
    from melony.core.result_backends import IResultBackend

__all__ = ()

type RedisCronConsumer = AsyncRedisCronConsumer | SyncRedisCronConsumer

_CRON_POP_SCRIPT = """ local items = redis.call('ZRANGEBYSCORE', KEYS[1], '-inf', ARGV[1])
if #items > 0 then
    redis.call('ZREM', KEYS[1], unpack(items))
end
return items
"""


@final
class AsyncRedisCronConsumer(BaseAsyncCronConsumer):
    def __init__(
        self,
        connection: Redis,
        broker: "BaseBroker",
        result_backend: "IResultBackend | None" = None,
    ) -> None:
        super().__init__(connection=connection, broker=broker, result_backend=result_backend)

    @override
    async def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
        now = datetime.now().timestamp()
        script = self._connection.register_script(_CRON_POP_SCRIPT)
        raw_items = await script(keys=[cron_queue], args=[str(now)])
        return [CronEntry.from_json(raw.decode("utf-8")) for raw in raw_items]

    @override
    async def _initialize_cron_tasks(self, cron_queue: str) -> None:
        now = datetime.now().timestamp()
        for reg in self._broker._cron_registry:
            if reg.queue == cron_queue:
                entry = CronEntry.from_registration(reg)
                next_run = croniter(reg.cron, now).get_next(float)
                await self._connection.zadd(  # noqa: WPS476
                    cron_queue,
                    {entry.as_json(): next_run},
                    nx=True,
                )

    @override
    async def _reschedule(self, entry: CronEntry, cron_queue: str) -> None:
        now = datetime.now().timestamp()
        next_run = croniter(entry.cron, now).get_next(float)
        await self._connection.zadd(cron_queue, {entry.as_json(): next_run})


@final
class SyncRedisCronConsumer(BaseSyncCronConsumer):
    def __init__(
        self,
        connection: SyncRedis,
        broker: "BaseBroker",
        result_backend: "IResultBackend | None" = None,
    ) -> None:
        super().__init__(connection=connection, broker=broker, result_backend=result_backend)

    @override
    def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
        now = datetime.now().timestamp()
        script = self._connection.register_script(_CRON_POP_SCRIPT)
        raw_items = script(keys=[cron_queue], args=[str(now)])
        return [CronEntry.from_json(raw.decode("utf-8")) for raw in raw_items]

    @override
    def _initialize_cron_tasks(self, cron_queue: str) -> None:
        now = datetime.now().timestamp()
        for reg in self._broker._cron_registry:
            if reg.queue == cron_queue:
                entry = CronEntry.from_registration(reg)
                next_run = croniter(reg.cron, now).get_next(float)
                self._connection.zadd(
                    cron_queue,
                    {entry.as_json(): next_run},
                    nx=True,
                )

    @override
    def _reschedule(self, entry: CronEntry, cron_queue: str) -> None:
        now = datetime.now().timestamp()
        next_run = croniter(entry.cron, now).get_next(float)
        self._connection.zadd(cron_queue, {entry.as_json(): next_run})



from typing import Awaitable, Final, Optional, Sequence, cast, final, override

from melony.core.task_converters import TaskConverter
from melony.core.brokers import BaseBroker
from melony.core.consumers import BaseAsyncConsumer, BaseSyncConsumer
from melony.core.publishers import IAsyncPublisher, ISyncPublisher
from melony.core.result_backends import IResultBackend
from melony.core.tasks import Task

__all__ = ()

type RedisConsumer = AsyncRedisConsumer | SyncRedisConsumer


@final
class AsyncRedisConsumer(BaseAsyncConsumer):
    _brpop_timeout: Final[float] = 0.1

    def __init__(
        self,
        publisher: IAsyncPublisher,
        broker: BaseBroker,
        result_backend: IResultBackend | None = None
    ) -> None:
        super().__init__(publisher, result_backend)
        self._broker = broker
        self._task_converter = TaskConverter()

    @override
    async def _pop_tasks(self, queue: str) -> Sequence[Task]:
        tasks: list[Task] = []
        redis_task = await cast(
            Awaitable[Sequence[bytes]],
            self._connection.brpop(keys=[queue])
        )
        tasks.append(self._deserialize_to_task_from_redis(redis_task))
        
        while redis_task:
            redis_task = await cast(
                Awaitable[Optional[Sequence[bytes]]],
                self._connection.brpop(
                    keys=[queue],
                    timeout=self._brpop_timeout
                )
            )
            if redis_task:
                tasks.append(self._deserialize_to_task_from_redis(redis_task))
        
        return tasks

    def _deserialize_to_task_from_redis(self, redis_task: Sequence[bytes]) -> Task:
        serialized_task_bytes = redis_task[1]
        serialized_task = serialized_task_bytes.decode("utf-8")
        task = self._task_converter.deserialize_task(
            serialized_task=serialized_task,
            broker=self._broker
        )
        return task


@final
class SyncRedisConsumer(BaseSyncConsumer):
    _brpop_timeout: Final[float] = 0.1

    def __init__(
        self,
        publisher: ISyncPublisher,
        broker: BaseBroker,
        result_backend: IResultBackend | None = None
    ) -> None:
        super().__init__(publisher, result_backend)
        self._broker = broker
        self._task_converter = TaskConverter()

    @override
    def _pop_tasks(self, queue: str) -> Sequence[Task]:
        tasks: list[Task] = []
        redis_task = cast(
            Sequence[bytes],
            self._connection.brpop(keys=[queue])
        )
        tasks.append(self._deserialize_to_task_from_redis(redis_task))
        
        while redis_task:
            redis_task = cast(
                Optional[Sequence[bytes]],
                self._connection.brpop(
                    keys=[queue],
                    timeout=self._brpop_timeout
                )
            )
            if redis_task:
                tasks.append(self._deserialize_to_task_from_redis(redis_task))
        
        return tasks

    def _deserialize_to_task_from_redis(self, redis_task: Sequence[bytes]) -> Task:
        serialized_task_bytes = redis_task[1]
        serialized_task = serialized_task_bytes.decode("utf-8")
        task = self._task_converter.deserialize_task(
            serialized_task=serialized_task,
            broker=self._broker
        )
        return task


import asyncio
import multiprocessing
import time

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Final, final
from redis.exceptions import ConnectionError

from melony.core.consts import CRON_QUEUE_PREFIX, DEFAULT_QUEUE
from melony.core.cron_tasks import CronEntry
from melony.core.task_finders import find_task_func
from melony.logger import log_error, log_info

if TYPE_CHECKING:
    from melony.core.brokers import BaseBroker
    from melony.core.result_backends import IResultBackend

__all__ = ()

_CRON_COMPLETE_MSG: Final[str] = "Cron task '{}' completed"
_CRON_FAILED_MSG: Final[str] = "Cron task '{}' failed on attempt {}"


class BaseCronConsumer:
    @final
    def _run_all_processes(self, processes: list) -> None:
        for process in processes:
            process.start()
        for process in processes:
            process.join()


class BaseAsyncCronConsumer(ABC, BaseCronConsumer):  # noqa: WPS214
    _poll_interval_secs: Final[int] = 1

    def __init__(
        self,
        connection,
        broker: "BaseBroker",
        result_backend: "IResultBackend | None" = None,
    ) -> None:
        self._connection = connection
        self._broker = broker
        self._result_backend = result_backend

    @final
    async def start_consume(
        self,
        queue: str = DEFAULT_QUEUE,
        processes: int = 1,
    ) -> None:
        cron_queue = f"{CRON_QUEUE_PREFIX}{queue}"
        if processes < 1:
            raise ValueError("Param 'processes' must be positive integer (without zero)")
        if processes == 1:
            await self._consumer_loop(cron_queue=cron_queue, consumer_id=0)
            return
        running_processes = []
        for process_num in range(processes):
            process = multiprocessing.Process(
                name=f"melony-cron-process-{process_num}",
                target=self._run_consumer_in_process,
                args=(cron_queue, process_num),
                daemon=False,
            )
            running_processes.append(process)
        try:
            self._run_all_processes(running_processes)
        except KeyboardInterrupt:
            for process in running_processes:
                if process.is_alive():
                    process.terminate()

    @final
    async def _consumer_loop(self, cron_queue: str, consumer_id: int = 0) -> None:
        await self._initialize_cron_tasks(cron_queue)
        log_info(f"Start listening cron queue {cron_queue}", consumer_id=consumer_id)
        while True:
            try:
                await self._consumer_loop_iteration(cron_queue, consumer_id)
            except ConnectionError as exc:
                log_error("Redis connection error", exc=exc)
                break
            except Exception as exc:
                log_error("Unexpected error at cron consuming loop", exc=exc)
                break
            await asyncio.sleep(self._poll_interval_secs)

    @final
    async def _consumer_loop_iteration(
        self,
        cron_queue: str,
        consumer_id: int,
    ) -> None:
        pop_due = await self._pop_due_entries(cron_queue)
        for entry in pop_due:
            await self._execute_entry(entry, cron_queue, consumer_id)  # noqa: WPS476

    @final
    async def _execute_entry(  # noqa: WPS231
        self,
        cron_entry: CronEntry,
        cron_queue: str,
        consumer_id: int,
    ) -> None:
        func = find_task_func(cron_entry.func_path)
        remaining = cron_entry.retries
        attempt_num = 0
        while remaining > 0:
            remaining -= 1
            attempt_num += 1
            try:
                if cron_entry.is_coro:
                    await func()
                else:
                    func()
            except Exception as exc:
                log_error(
                    _CRON_FAILED_MSG.format(cron_entry.func_path, attempt_num),
                    exc=exc,
                    consumer_id=consumer_id,
                )
                if remaining > 0:
                    await asyncio.sleep(cron_entry.retry_timeout)
                continue
            log_info(_CRON_COMPLETE_MSG.format(cron_entry.func_path), consumer_id=consumer_id)
            break
        await self._reschedule(cron_entry, cron_queue)

    @final
    def _run_consumer_in_process(self, cron_queue: str, consumer_id: int = 0) -> None:
        asyncio.run(self._consumer_loop(cron_queue=cron_queue, consumer_id=consumer_id))

    @abstractmethod
    async def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
        """Atomically pop all due cron entries from the sorted set."""

    @abstractmethod
    async def _initialize_cron_tasks(self, cron_queue: str) -> None:
        """Initialize cron tasks."""

    @abstractmethod
    async def _reschedule(self, entry: CronEntry, cron_queue: str) -> None:
        """Reschedule a cron entry with its next execution time."""


class BaseSyncCronConsumer(ABC, BaseCronConsumer):  # noqa: WPS214
    _poll_interval_secs: Final[int] = 1

    def __init__(
        self,
        connection,
        broker: "BaseBroker",
        result_backend: "IResultBackend | None" = None,
    ) -> None:
        self._connection = connection
        self._broker = broker
        self._result_backend = result_backend

    @final
    def start_consume(
        self,
        queue: str = DEFAULT_QUEUE,
        processes: int = 1,
    ) -> None:
        cron_queue = f"{CRON_QUEUE_PREFIX}{queue}"
        if processes < 1:
            raise ValueError("Param 'processes' must be positive integer (without zero)")
        if processes == 1:
            self._consumer_loop(cron_queue=cron_queue, consumer_id=0)
            return
        running_processes = []
        for process_num in range(processes):
            process = multiprocessing.Process(
                name=f"melony-cron-process-{process_num}",
                target=self._consumer_loop,
                args=(cron_queue, process_num),
                daemon=False,
            )
            running_processes.append(process)
        try:
            self._run_all_processes(running_processes)
        except KeyboardInterrupt:
            for process in running_processes:
                if process.is_alive():
                    process.terminate()

    @final
    def _consumer_loop(self, cron_queue: str, consumer_id: int = 0) -> None:
        self._initialize_cron_tasks(cron_queue)
        log_info(f"Start listening cron queue {cron_queue}", consumer_id=consumer_id)
        while True:
            try:
                self._consumer_loop_iteration(cron_queue, consumer_id)
            except ConnectionError as exc:
                log_error("Redis connection error", exc=exc)
                break
            except Exception as exc:
                log_error("Unexpected error at cron consuming loop", exc=exc)
                break
            time.sleep(self._poll_interval_secs)

    @final
    def _consumer_loop_iteration(
        self,
        cron_queue: str,
        consumer_id: int,
    ) -> None:
        pop_due = self._pop_due_entries(cron_queue)
        for entry in pop_due:
            self._execute_entry(entry, cron_queue, consumer_id)

    @final
    def _execute_entry(
        self,
        entry: CronEntry,
        cron_queue: str,
        consumer_id: int,
    ) -> None:
        func = find_task_func(entry.func_path)
        remaining = entry.retries
        attempt_num = 0
        while remaining > 0:
            remaining -= 1
            attempt_num += 1
            try:
                func()
            except Exception as exc:
                log_error(
                    _CRON_FAILED_MSG.format(entry.func_path, attempt_num),
                    exc=exc,
                    consumer_id=consumer_id,
                )
                if remaining > 0:
                    time.sleep(entry.retry_timeout)
                continue
            log_info(_CRON_COMPLETE_MSG.format(entry.func_path), consumer_id=consumer_id)
            break
        self._reschedule(entry, cron_queue)

    @final
    def _run_consumer_in_process(self, cron_queue: str, consumer_id: int = 0) -> None:
        self._consumer_loop(cron_queue=cron_queue, consumer_id=consumer_id)

    @abstractmethod
    def _pop_due_entries(self, cron_queue: str) -> list[CronEntry]:
        """Atomically pop all due cron entries from the sorted set."""

    @abstractmethod
    def _initialize_cron_tasks(self, cron_queue: str) -> None:
        """Initialize cron tasks."""

    @abstractmethod
    def _reschedule(self, entry: CronEntry, cron_queue: str) -> None:
        """Reschedule a cron entry with its next execution time."""



from setuptools import setup, find_packages


def readme():
  with open("README.md", "r") as file:
    return file.read()


setup(
  name="melony",
  version="1.3.0",
  author="wimble3",
  author_email="wimble@internet.ru",
  description='Multisync task manager for python',
  long_description=readme(),
  long_description_content_type="text/markdown",
  url="https://melony-python.com/",
  packages=find_packages(),
  install_requires=[
    "asyncio>=4.0.0",
    "classes>=0.4.1",
    "pytest>=8.4.1",
    "pytest-asyncio>=1.1.0",
    "redis>=6.4.0",
    "setuptools>=80.10.2",
    "wemake-python-styleguide>=1.3.0",
  ],
  classifiers=[
    "Programming Language :: Python :: 3.12",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent"
  ],
  keywords='task manager python melony',
  project_urls={
    "Documentation": "https://melony-python.com/docs"
  },
  python_requires='>=3.12'
)



import pytest
from unittest.mock import MagicMock

from melony.core.brokers import BaseBroker
from melony.core.task_wrappers import AsyncTaskWrapper, SyncTaskWrapper


class _StubBroker(BaseBroker):
    @property
    def publisher(self):
        return MagicMock()

    @property
    def consumer(self):
        return MagicMock()

    @property
    def cron_consumer(self):
        return MagicMock()


def test_cron_task_registered_in_registry():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    def my_task():
        pass

    assert len(broker._cron_registry) == 1


def test_cron_task_stores_correct_cron_expression():
    broker = _StubBroker()

    @broker.task(cron="0 8 * * 1-5")
    def my_task():
        pass

    assert broker._cron_registry[0].cron == "0 8 * * 1-5"


def test_cron_task_is_coro_false_for_sync():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    def sync_task():
        pass

    assert broker._cron_registry[0].is_coro is False


def test_cron_task_is_coro_true_for_async():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    async def async_task():
        pass

    assert broker._cron_registry[0].is_coro is True


def test_cron_task_default_queue():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    def my_task():
        pass

    assert broker._cron_registry[0].queue == "melony_cron:default"


def test_cron_task_custom_queue():
    broker = _StubBroker()

    @broker.task(cron="* * * * *", queue="reports")
    def my_task():
        pass

    assert broker._cron_registry[0].queue == "melony_cron:reports"


def test_cron_task_stores_retries_and_timeout():
    broker = _StubBroker()

    @broker.task(cron="* * * * *", retries=5, retry_timeout=30)
    def my_task():
        pass

    reg = broker._cron_registry[0]
    assert reg.retries == 5
    assert reg.retry_timeout == 30


def test_cron_task_stores_original_func():
    broker = _StubBroker()

    def original():
        return "result"

    broker.task(cron="* * * * *")(original)

    assert broker._cron_registry[0].func() == "result"


def test_cron_task_func_path_contains_qualname():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    def my_task():
        pass

    assert "my_task" in broker._cron_registry[0].func_path


def test_multiple_cron_tasks_all_registered():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    def task_one():
        pass

    @broker.task(cron="0 12 * * *")
    def task_two():
        pass

    assert len(broker._cron_registry) == 2
    crons = {r.cron for r in broker._cron_registry}
    assert crons == {"* * * * *", "0 12 * * *"}


def test_no_cron_does_not_register():
    broker = _StubBroker()

    @broker.task
    def my_task():
        pass

    assert len(broker._cron_registry) == 0


def test_no_cron_with_params_does_not_register():
    broker = _StubBroker()

    @broker.task(retries=3)
    def my_task():
        pass

    assert len(broker._cron_registry) == 0


def test_cron_sync_task_still_returns_sync_wrapper():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    def my_task():
        pass

    assert isinstance(my_task(), SyncTaskWrapper)


def test_cron_async_task_still_returns_async_wrapper():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    async def my_task():
        pass

    assert isinstance(my_task(), AsyncTaskWrapper)


def test_invalid_cron_raises_value_error():
    broker = _StubBroker()

    with pytest.raises(ValueError, match="Invalid cron expression"):
        @broker.task(cron="not-a-cron")
        def my_task():
            pass


def test_invalid_cron_five_asterisks_is_valid():
    broker = _StubBroker()

    @broker.task(cron="* * * * *")
    def my_task():
        pass

    assert len(broker._cron_registry) == 1


def test_zero_retries_raises_value_error():
    broker = _StubBroker()

    with pytest.raises(ValueError, match="retries"):
        @broker.task(cron="* * * * *", retries=0)
        def my_task():
            pass


def test_negative_retry_timeout_raises_value_error():
    broker = _StubBroker()

    with pytest.raises(ValueError, match="retry_timeout"):
        @broker.task(cron="* * * * *", retry_timeout=-1)
        def my_task():
            pass



import asyncio

from typing import Any, final, Iterable
from redis.asyncio import Redis
from redis import Redis as SyncRedis

from melony.core.dto import TaskResultDTO, TaskExecResultsDTO
from melony.core.result_backends import IAsyncResultBackendSaver, ISyncResultBackendSaver
from melony.core.tasks import Task
from melony.logger import log_error, log_info

__all__ = ()


@final
class AsyncTaskExecutor:
    def __init__(
        self,
        connection: Redis,
        result_backend_saver: IAsyncResultBackendSaver | None = None
    ) -> None:
        self._connection = connection
        self._result_backend = result_backend_saver

    @final
    async def execute_tasks(
        self,
        tasks: Iterable[Task],
        consumer_id: int
    ) -> TaskExecResultsDTO:
        task_map: dict[asyncio.Task, Task] = {}
        asyncio_tasks: list[asyncio.Task] = []
        
        for task in tasks:
            asyncio_task = asyncio.create_task(task.execute())
            task_map[asyncio_task] = task
            asyncio_tasks.append(asyncio_task)
            if task._meta.retries_left:
                task._meta.retries_left -= 1
        
        tasks_exec_results = await self._asyncio_wait_tasks(
            pending=asyncio_tasks,
            task_map=task_map,
            consumer_id=consumer_id
        )
        return tasks_exec_results

    @final
    async def _asyncio_wait_tasks(  # noqa: WPS210, WPS231
        self,
        pending: Iterable[asyncio.Task],
        task_map: dict[asyncio.Task, Task],
        consumer_id: int
    ) -> TaskExecResultsDTO:
        tasks_to_retry: list[Task] = []
        tasks_done: list[TaskResultDTO] = []
        while pending:
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )
            for completed_asyncio_task in done:
                task = task_map[completed_asyncio_task]
                task_result = self._get_task_result(asyncio_task=completed_asyncio_task)
                if isinstance(task_result, BaseException):
                    log_error(
                        f"Task with id {task.task_id} has been executed with error: "
                        f"{task_result}",
                        exc=task_result,
                        consumer_id=consumer_id
                    )
                    tasks_to_retry.append(task)
                    continue
                log_info(
                    f"Task with id '{task.task_id}' completed with result "
                    f"{task_result=}",
                    consumer_id=consumer_id
                )
                tasks_done.append(TaskResultDTO(task=task, task_result=task_result))
            if self._result_backend:
                await self._result_backend.save_results(task_results=tasks_done)
        return TaskExecResultsDTO(
            tasks_with_result=tasks_done,
            tasks_to_retry=tasks_to_retry
        )

    @final
    def _get_task_result(self, asyncio_task: asyncio.Task) -> Any:  # TODO: typing
        task_exc = asyncio_task.exception()
        if task_exc:
            return task_exc
        task_result = asyncio_task.result()
        return task_result


@final
class SyncTaskExecutor:
    def __init__(
        self,
        connection: SyncRedis,
        result_backend_saver: ISyncResultBackendSaver | None = None
    ) -> None:
        self._connection = connection
        self._result_backend_saver = result_backend_saver

    @final
    def execute_tasks(
        self,
        tasks: Iterable[Task], 
        consumer_id: int
    ) -> TaskExecResultsDTO:
        tasks_to_retry: list[Task] = []
        tasks_done: list[TaskResultDTO] = []
        for task in tasks:
            if task._meta.retries_left:
                task._meta.retries_left -= 1
            try:
                task_result = task.execute()
            except Exception as exc:
                task_result = exc

            if isinstance(task_result, BaseException):
                log_error(
                    f"Task with id '{task.task_id}' has been executed with error: "
                    f"{task_result}",
                    exc=task_result
                )
                tasks_to_retry.append(task)
                continue

            log_info(
                f"Task with id '{task.task_id}' completed with result {task_result=}",
                consumer_id=consumer_id
            )
            tasks_done.append(TaskResultDTO(task=task, task_result=task_result))

            if self._result_backend_saver:
                self._result_backend_saver.save_results(task_results=tasks_done)
        
        return TaskExecResultsDTO(
            tasks_with_result=tasks_done,
            tasks_to_retry=tasks_to_retry
        )



import importlib

from typing import Callable

__all__ = ()


def find_task_func(func_path: str) -> Callable:
    try:
        return _find_func(func_path)
    except (ImportError, AttributeError, ValueError) as exc:
        raise ImportError(f"Cannot import function '{func_path}': {exc}")


def _find_func(func_path: str) -> Callable:
    module_name, func_name = func_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    func = getattr(module, func_name)
    
    if not callable(func):
        raise ValueError(f"'{func_name}' is not callable")
        
    return func


from typing import Awaitable, cast, final, override
from redis.asyncio import Redis
from redis import Redis as SyncRedis

from melony.core.publishers import IAsyncPublisher, ISyncPublisher
from melony.core.tasks import Task
from melony.core.task_converters import TaskConverter

__all__ = ()


@final
class AsyncRedisPublisher(IAsyncPublisher):
    def __init__(self, connection: Redis) -> None:
        self._connection = connection
        self._task_converter = TaskConverter()

    @property
    @override
    def connection(self) -> Redis:
        return self._connection

    @override
    async def push(self, task: Task) -> None:
        await cast(
            Awaitable[int], self._connection.lpush(
                task.queue,
                self._task_converter.serialize_task(task)
            )
        )


@final
class SyncRedisPublisher(ISyncPublisher):
    def __init__(self, connection: SyncRedis) -> None:
        self._connection = connection
        self._task_converter = TaskConverter()

    @property
    @override
    def connection(self) -> SyncRedis:
        return self._connection

    @override
    def push(self, task: Task) -> None:
        self._connection.lpush(
            task.queue,
            self._task_converter.serialize_task(task)
        )



from abc import ABC, abstractmethod
from typing import Iterable

from melony.core.dto import TaskResultDTO

__all__ = ()

type ResultBackendSaver = IAsyncResultBackendSaver | ISyncResultBackendSaver


class IAsyncResultBackendSaver(ABC):
    @abstractmethod
    async def save_results(self, task_results: Iterable[TaskResultDTO]) -> None:
        ...

class ISyncResultBackendSaver(ABC):
    @abstractmethod
    def save_results(self, task_results: Iterable[TaskResultDTO]) -> None:
        ...

class IResultBackend(ABC):
    @property
    @abstractmethod
    def saver(self) -> ResultBackendSaver:
        ...
